#include <assert.h>
#include <allegro.h>
#include <winalleg.h>
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <winsock.h>
#include <dirent.h>
#include "grain_engine.h"


#ifndef M_PI
#define M_PI 3.1415926535897932384338
#endif


#define sqr(X) ((X)*(X))
typedef struct ball_struct
{
  double x, y, z;
  double dx, dy, dz;
  double ddx, ddy, ddz;
  double *sx, *sy, *sz;
  double px, py, pz;
} ball;


typedef struct roulette_struct
{
  ball ***bearings;
  int steps, branch, horizon, sens;
} roulette;

typedef struct surface_struct
{
  int *surface;
  int *noise;
  double *dx, *dy;
  int s_x, s_y;
  double friction, wall_bounce;
} surface;

FILE *marker_file = NULL;
roulette *roulette_sim;
ball *one_bearing;
static double tilt_angle_x, tilt_angle_y, tilt_angle_z;
static double mtilt_angle_x, mtilt_angle_y, mtilt_angle_z;

static int monte_carlo = 0;

static int screen_w, screen_h;
static surface *ground, *bike_ground;
static double petrb = 0.1, pos_petrb = 0.8, step_ptrb = 0.7;
static int shadowing = 0, trans_mode = 1;
static double heading = 0.0;
static double sscale = 1.0;
static int jump_to_arrow = 0;
#define SAMPLE_RATE 44100

static int BUFSIZE = 4096;
static int QUALITY = 2;

static BITMAP *screen_buf;
BITMAP *map, *shadow, *bike;
static AUDIOSTREAM *out_stream;

static int n_waves;
static char **wave_names;
static char **marker_wave_names;
static int *wave_x, *wave_y;

static double width = 15.0;
static int phase = 0;
static double rate = 1.0;
static int ego = 1, egopos = 1;
static int drop_mode = 0, cur_wave = 0;




void screen_pos (double x, double y, double *ox, double *oy)
{
double nx, ny;
if(egopos)
{
 nx = (x-one_bearing->x) * sscale ;
 ny = (y-one_bearing->z) * sscale;
}
else
{
nx = x * sscale ;
 ny = y * sscale;

}
 if(ego)
 {
 *ox = cos(-heading) * nx + sin(-heading) * ny + SCREEN_W/2;
 *oy = sin(-heading) * nx - cos(-heading) * ny + SCREEN_H/2;
 }
 else
 {
 *ox = nx + SCREEN_W/2;
 *oy = ny + SCREEN_H/2;
 }
}





static void
read_wave_names ()
{
  int i;

  DIR *directory;
  struct dirent *de;
  directory = opendir (".");
  while ((de = readdir (directory)))
    if (de->d_namlen > 4
	&& !strcasecmp (de->d_name + (de->d_namlen - 4), ".wav"))
      n_waves++;
  rewinddir (directory);
  wave_names = malloc (sizeof (*wave_names) * n_waves);

  assert (wave_names != NULL);

  i = 0;
  while ((de = readdir (directory)))
    {
      if (de->d_namlen > 4
	  && !strcasecmp (de->d_name + (de->d_namlen - 4), ".wav"))
	{
	  wave_names[i] = malloc (sizeof (char) * de->d_namlen + 1);
	  strcpy (wave_names[i], de->d_name);
	  i++;
	}
    }
  closedir (directory);

}




int read_marker_line(FILE *infile, char *outname, int *outx, int *outy)
{
	char line[2048];
	char *result, *endptr;
	char *token;
	if(infile==NULL)
		return 0;
	result = fgets(line, 2047, infile);
	if(!result)
		return 0;
	token = strtok(line, ",");
	strcpy(outname, token);
	token = strtok(NULL, ",");
	*outx = strtol(token, &endptr, 10);
	token = strtok(NULL, ",\n ");
	*outy = strtol(token, &endptr, 10);

	

}

void add_wave(char *name, int x, int y, int i)
{  
grain_source *source;
  ex_position_2d *pos;
  SAMPLE *samp;
	  marker_wave_names[i] = strdup(name);
      samp = load_sample (name);
      if (samp != NULL)
		{
		source = create_uniform_source (i, samp->data, samp->len * 2);
		resample_wave (source, QUALITY);
		pos = calloc (sizeof (*pos), 1);
		assert (pos != NULL);
		pos->x = x;
	    pos->y = y;
	    wave_x[i] = pos->x;
	    wave_y[i] = pos->y;
	    set_extended_properties (source, EX_POSITION_2D, pos);
	    add_source (source);
	    }
      else
	printf ("Wave %s was NULL\n", wave_names[i]);

}

void
load_waveforms ()
{
  
  FILE *markers;


  char outname[2048];
  int outx, outy, read_line;
  int i;
  read_wave_names ();
  marker_wave_names = calloc(sizeof(*marker_wave_names), n_waves*2);
  markers = fopen("markers.txt", "r");
  
  wave_x = calloc (sizeof (*wave_x), n_waves);
  wave_y = calloc (sizeof (*wave_y), n_waves);

  assert (wave_x != NULL);
  assert (wave_y != NULL);
  srand (time (NULL));
  for (i = 0; i < n_waves; i++)
    {
	  read_line = read_marker_line(markers, outname, &outx, &outy);
	  
	  if(read_line)
		add_wave(outname, outx, outy, i);
	  else
		add_wave(wave_names[i], 0, 0, i);
		
	  
    }
	if(markers)
	 fclose(markers);

}

double
distance (double x1, double y1, double x2, double y2)
{
  return sqrt ((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
}



void
play_audio ()
{
  unsigned char *buf;
  long rest_time;
  int i;

  out_stream =
    play_audio_stream (BUFSIZE, 16, FALSE, SAMPLE_RATE / QUALITY, 255, 128);

  set_buffer_length (128);
  rest_time = 1;
  while (key[KEY_ESC] == 0)
    {
      buf = get_audio_stream_buffer (out_stream);
      if (buf == NULL)
	rest (rest_time);
      else
	{
	  fill_buffer (buf, BUFSIZE * 2);
	  free_audio_stream_buffer (out_stream);
	}
    }
  stop_audio_stream (out_stream);
}

double cur_pitch = 1.0;

double
update_source_pdf (grain_source * source)
{
  
  int dx, dy, i;
  double dist;
  double dt;
  double val, accum;
  int mc_ctr;
  ball *bearing;
  ex_position_2d *pos;
  
  
  if (get_extended_type (source) == EX_POSITION_2D)
    pos = (ex_position_2d *) (get_extended_properties (source));
  else
    return 1.0;
  accum = 0.0;
  
  for(i=0;i<100;i++)
  {
    mc_ctr = rand() % roulette_sim->branch;
	bearing = roulette_sim->bearings[roulette_sim->horizon - 1][mc_ctr];
	dx = pos->x - bearing->x;
	dy = pos->y - bearing->z;

     dist = sqrt (dx * dx + dy * dy);
     val = (exp (-dist / (2 * (width / 5.0) * (width / 5.0))));
	 accum += val;
   }
  dx = pos->x - one_bearing->x;
  dy = pos->y - one_bearing->z;
  
  dist = sqrt(dx*dx + dy * dy);
  dt = 12000.0/(dist+50);
  if(dt>150)
	dt = 150;
  source->time_position += dt;
  set_probability (source, accum/100.0);
  set_position(source, source->time_position, dist*10+50);

  return val;
}

void
init_grains ()
{
  set_format (SAMPLE_RATE / QUALITY, 16, 1);
  set_max_grains (1000);
  set_envelope (ENV_LINEAR, 0.05);

  set_normalize (0);
  set_grain_length (10000);
  register_distribution_function (update_source_pdf);
  initialize_engine ();

}


#define rand_double() ((double)rand() / (double)RAND_MAX)

static double
rand_gaussian ()
{
  static int precomputed_next = 0;
  static double next = 0.0;
  double multiplier;
  if (precomputed_next)
    {
      precomputed_next = 0;
      return next;
    }
  else
    {
      double v1, v2, s;
      do
	{
	  v1 = 2 * rand_double () - 1;
	  v2 = 2 * rand_double () - 1;
	  s = v1 * v1 + v2 * v2;
	}
      while (s >= 1 || s == 0);

      multiplier = sqrt (-2 * log (s) / s);
      next = v2 * multiplier;
      precomputed_next = 1;
      return v1 * multiplier;
    }
}

int get_dx(surface *surf, int bx, int bz)
{
int s_x, s_y;
  s_x = surf->s_x;
  s_y = surf->s_y;

  if (fabs (bx) >= s_x-1 || fabs (bz) >= s_y-1)
    return 0.0;
  bx += s_x;
  bz+= s_y;
	return surf->dx[bx+bz*s_x*2];

}

int get_dy(surface *surf, int bx, int bz)
{
int s_x, s_y;
  s_x = surf->s_x;
  s_y = surf->s_y;

  if (fabs (bx) >= s_x-1 || fabs (bz) >= s_y-1)
    return 0.0;
  bx += s_x;
  bz+= s_y;
	return surf->dy[bx+bz*s_x*2];

}


int surf_height(surface *surf, int bx, int bz)
{
int s_x, s_y;
  s_x = surf->s_x;
  s_y = surf->s_y;

  if (fabs (bx) >= s_x-1 || fabs (bz) >= s_y-1)
    return 0.0;
  bx += s_x;
  bz+= s_y;
	return surf->surface[bx+bz*s_x*2];
}

int get_noise(surface *surf, int bx, int bz)
{
int s_x, s_y;
  s_x = surf->s_x;
  s_y = surf->s_y;

  if (fabs (bx) >= s_x-1 || fabs (bz) >= s_y-1)
    return 0.0;
 bx += s_x;
  bz+= s_y;
		return surf->noise[bx+bz*s_x*2];
}



ball *
create_ball (double x, double y, double z)
{
  ball *bearing;
  bearing = calloc (sizeof (*bearing), 1);
  assert (bearing != NULL);
  bearing->x = 0;
  bearing->y = 0;
  bearing->z = 0;
  bearing->dx = 0;
  bearing->dy = 0;
  bearing->dz = 0;
  bearing->ddx = 0;
  bearing->ddy = 0;
  bearing->ddz = 0;
  bearing->px = 0;
  bearing->py = 0;
  bearing->pz = 0;

  bearing->sx = calloc (sizeof (*(bearing->sx)), 2);

  bearing->sy = calloc (sizeof (*(bearing->sy)), 2);
  bearing->sz = calloc (sizeof (*(bearing->sz)), 2);
  assert (bearing->sx != NULL);
  assert (bearing->sx != NULL);
  assert (bearing->sx != NULL);
  return bearing;
}

surface *
create_surface (BITMAP *h, BITMAP *n)
{
  surface *surf;
  int i,j,k;
  double x, x1, x2, y, y1, y2;
  surf = calloc (sizeof (*surf), 1);
  assert (surf != NULL);
  surf->s_x = h->w/2;
  surf->s_y = h->h/2;
  surf->friction = 0.5;
  surf->wall_bounce = 0.1;
  surf->surface = calloc (sizeof (*(surf->surface)), 4 * surf->s_x * surf->s_y);
  surf->dx = calloc (sizeof (*(surf->dx)), 4 * surf->s_x * surf->s_y);
  surf->dy = calloc (sizeof (*(surf->dy)), 4 * surf->s_x * surf->s_y);
  surf->noise = calloc (sizeof (*(surf->noise)), 4 * surf->s_x * surf->s_y);
  for(i=0;i<h->w;i++)
	for(j=0;j<h->h;j++)
		{
		    k = j;
			surf->surface[i+k*h->w] = getpixel(h, i, j)/1.6e4;
			surf->surface[i+k*h->w] *= surf->surface[i+k*h->w]/10.0;
			
			surf->noise[i+k*n->w] = (60.0 - getpixel(n, i, j)/1.6e5)/2.0;
			
		}
  for(i=1;i<h->w-1;i++)
	for(j=1;j<h->h-1;j++)
		{
		    k = j;
			x = getpixel(h, i, j)/1.6e4;
			x1 = getpixel(h, i-1, j)/1.6e4;
			x2 = getpixel(h, i+1, j)/1.6e4;
			
			y1 = getpixel(h, i, j-1)/1.6e4;
			y2 = getpixel(h, i, j+1)/1.6e4;
			
			surf->dx[i+k*h->w] = -atan2(((x-x1)+(x2-x))/3,1.0);
			surf->dy[i+k*h->w] = -atan2(((x-y1)+(y2-x))/3,1.0);
			
		}
  
  
  assert (surf->surface != NULL);
  return surf;
}



void
updateDerivs (surface * surf, ball * bearing)
{
  double dx, dz;
  double tx, tz;
  double dt = 1.5;

  dx = get_dx(surf, bearing->x, bearing->z);
  dz = get_dy(surf, bearing->x, bearing->z);
    
   bearing->ddx = (sin (tilt_angle_x + dx) );
    
   bearing->ddz = (sin (tilt_angle_z + dz));
      
    
  bearing->ddx -= surf->friction * bearing->dx;
  bearing->ddz -= surf->friction * bearing->dz;


  bearing->dx = (bearing->dx + bearing->ddx * dt);
  bearing->dy = (bearing->dy + bearing->ddy * dt);
  bearing->dz = (bearing->dz + bearing->ddz * dt);



  bearing->x += bearing->dx * dt;
  bearing->y += bearing->dy * dt;
  bearing->z += bearing->dz * dt;
  
  if (bearing->x  >= surf->s_x)
		bearing->x = surf->s_x-1;
  if (bearing->x  <= -surf->s_x)
		bearing->x = (-surf->s_x)+1;
  if (bearing->z  <= -surf->s_y) 
		bearing->z = (-surf->s_y)+1;
  if (bearing->z  >= surf->s_y)
		bearing->z = surf->s_y-1;


}



void
updateRoulette (void)
{
  int i, j;
  int steps, branch, sens;
  int n_balls;

  double old_tx, old_tz;
  double old_x, old_z;
  double ptrb;
  old_tx = tilt_angle_x;
  old_tz = tilt_angle_z;
  old_x = one_bearing->x;
  old_z = one_bearing->z;
  steps = roulette_sim->steps;
  branch = roulette_sim->branch;
  sens = roulette_sim->sens;
  for (i = 0; i < branch; i++)
    {
	  ptrb = get_noise(ground, one_bearing->x, one_bearing->z);
      one_bearing->x = old_x + rand_gaussian () * ptrb * pos_petrb;
      one_bearing->z = old_z + rand_gaussian () * ptrb * pos_petrb;

      for (j = 0; j < sens; j++)
	{
	  
	  tilt_angle_x = old_tx + rand_gaussian () * petrb;
	  tilt_angle_z = old_tz + rand_gaussian () * petrb;
	  one_bearing->px = tilt_angle_x - old_tx;
	  one_bearing->pz = tilt_angle_z - old_tz;
	  peturb (i * sens + j, sens * branch, one_bearing, 
		  tilt_angle_x, tilt_angle_z, 0, roulette_sim->horizon);
	}
    }
  one_bearing->x = old_x;
  one_bearing->z = old_z;
  tilt_angle_x = old_tx;
  tilt_angle_z = old_tz;

}


void
copy_ball (ball * new_ball, ball * orig)
{
  new_ball->x = orig->x;
  new_ball->y = orig->y;
  new_ball->z = orig->z;
  new_ball->dx = orig->dx;
  new_ball->dy = orig->dy;
  new_ball->dz = orig->dz;
  new_ball->ddx = orig->ddx;
  new_ball->ddy = orig->ddy;
  new_ball->ddz = orig->ddz;
  new_ball->px = orig->px;
  new_ball->py = orig->py;
  new_ball->pz = orig->pz;

}




void
peturb (int branch, int n_branch, ball * orig,
	double old_tx, double old_tz, int step, int max_step)
{
  double ptrb;
  ball *new_ball;
  int i;

  if (step == max_step)
    return;

  new_ball = roulette_sim->bearings[step][branch];

  copy_ball (new_ball, orig);
  
  new_ball->x += rand_gaussian() * step_ptrb;
  new_ball->z += rand_gaussian() * step_ptrb;
 
  updateDerivs (ground, new_ball);

  peturb (branch, n_branch, new_ball, 
	  tilt_angle_x, tilt_angle_z, step + 1, max_step);

}

void
arrow (double x, double y, double tx, double tz, int col)
{
  double head_len = 0.3;
  double head_width = 0.3;
  double grad, inv_grad, pt_x, pt_z;
  double ang;
  double len = sqrt (tx * tx + tz * tz);
  double ox, oy, ox2, oy2;
  head_width = len / 10.0;

  pt_x = (1.0 - head_len) * tx;
  pt_z = (1.0 - head_len) * tz;

  ang = atan2 (tz, tx) - 0.5 * M_PI;
  
  screen_pos(x,y,&ox,&oy);
  screen_pos(x+tx, y+tz, &ox2, &oy2);
  line (screen_buf, ox, oy, ox2, oy2, col);
  if(jump_to_arrow)
  {
	jump_to_arrow = 0;
	one_bearing->x = x+tx;
	one_bearing->z = y+tz;
  }
  
  screen_pos(x+pt_x+head_width*cos(ang),y+pt_z+head_width*sin(ang),&ox,&oy);
  
  
  line (screen_buf, ox, oy, ox2, oy2, col);
  screen_pos(x+pt_x-head_width*cos(ang),y+pt_z-head_width*sin(ang),&ox,&oy);
  
  line (screen_buf, ox, oy, ox2, oy2, col);

}


void
drawRoulette (void)
{
  int i, j, n_balls;
  double radius;
  double p_scale = 5.0;
  double h;
  double ox, oy, ox2, oy2;
  ball *bearing;

  n_balls = roulette_sim->branch * roulette_sim->sens;

 screen_pos(one_bearing->x, one_bearing->z, &ox2, &oy2);
  
if(trans_mode)
{
 drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	set_trans_blender (128, 128, 128, 20);
}
j = roulette_sim->horizon -1;
  for (i = 0; i < roulette_sim->branch; i++)
    {
      double ox, oy, oz;
      bearing = roulette_sim->bearings[j][i];
      ox = bearing->x;
      oy = bearing->y;
      oz = bearing->z;

      for (j = 0; j < roulette_sim->sens; j++)
	{
	  bearing =
	    roulette_sim->bearings[roulette_sim->horizon -
				   1][i * roulette_sim->sens + j];

	  screen_pos(bearing->x, bearing->z, &ox, &oy);
	 
	  circlefill (screen_buf, ox, oy, 2, makecol (100, 255, 100));


	}

    }
	solid_mode ();
}





void draw_maps(void)
{
	fixed ang;
	double ox, oy;
	BITMAP *draw_map;
	
	if(shadowing)
		draw_map = shadow;
	else
		draw_map = map;
		
	
	ang = -heading/2*M_PI*207.0*8192.0;
	if(!ego)
	{
	screen_pos(0,0,&ox,&oy);
		pivot_scaled_sprite(screen_buf, draw_map,  ox, oy, draw_map->w/2,
	draw_map->h/2,
	 0, ftofix(sscale));
	}
	else
	{
	screen_pos(0,0,&ox,&oy);
	pivot_scaled_sprite_v_flip(screen_buf, draw_map,  ox, oy, draw_map->w/2,
	draw_map->h/2,
	 ang, ftofix(sscale));
	}
}

void show_stats(void)
{
char stats[1024];
sprintf(stats, "Branches:%d Steps:%d Total particles %d", roulette_sim->branch,
roulette_sim->steps, roulette_sim->steps*roulette_sim->branch);
textout(screen_buf,  font, stats, 10, 10, makecol(128,128,128));
sprintf(stats, "Shadow factor:%f Pos dispr.:%f D dispr. %f", pos_petrb,
step_ptrb, petrb);
textout(screen_buf,  font, stats, 10, 20, makecol(128,128,128));

if(drop_mode)
	{
		sprintf(stats, "Current Wave File: %s", marker_wave_names[cur_wave]);
		textout(screen_buf,  font, stats, 10, 30, makecol(255,128,128));
	}
}

void
display (void)
{
  int s_x, s_y;
  int i;
  double ox, oy, ox2, oy2;
  clear (screen_buf);

  tilt_angle_x = mtilt_angle_x;
  tilt_angle_z = mtilt_angle_z;



  if (monte_carlo)
    updateRoulette ();


  s_x = ground->s_x;
  s_y = ground->s_y;

  draw_maps();
  
  screen_pos(-s_x, -s_y, &ox, &oy);
  screen_pos(s_x, -s_y, &ox2, &oy2);
  line(screen_buf, ox, oy, ox2, oy2, makecol(255,255,255));
  screen_pos(s_x, -s_y, &ox, &oy);
  screen_pos(s_x, s_y, &ox2, &oy2);
  line(screen_buf, ox, oy, ox2, oy2, makecol(255,255,255));
  screen_pos(s_x, s_y, &ox, &oy);
  screen_pos(-s_x, s_y, &ox2, &oy2);
  line(screen_buf, ox, oy, ox2, oy2, makecol(255,255,255));
  screen_pos(-s_x, s_y, &ox, &oy);
  screen_pos(-s_x, -s_y, &ox2, &oy2);
  line(screen_buf, ox, oy, ox2, oy2, makecol(255,255,255));
  
  

screen_pos(one_bearing->x, one_bearing->z, &ox, &oy);
  circlefill (screen_buf, ox, oy, 5, makecol (100, 150, 150));

  arrow (one_bearing->x, one_bearing->z,
	 tilt_angle_x * roulette_sim->horizon *2 ,
	 tilt_angle_z * roulette_sim->horizon *2, makecol (100, 100, 230));

  drawRoulette ();

  for (i = 0; i < n_waves; i++)
    {
	  screen_pos(wave_x[i], wave_y[i], &ox, &oy);
      circlefill (screen_buf, ox, oy, 3, makecol (255, 0, 0));


    }
	
	
  show_stats();
  blit (screen_buf, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);



}


static unsigned char *buf, *dbl_buf;

void
timerFunc (void)
{
  static int copied = 0;
  motionFunc (mouse_x, mouse_y);
  buf = get_audio_stream_buffer (out_stream);
  if (buf == NULL)
    {
      if (!copied)
	{
	  fill_buffer (dbl_buf, BUFSIZE * 2);
	  copied = 1;
	}
    }
  else
    {
      memcpy (buf, dbl_buf, BUFSIZE * 2);
      copied = 0;
      free_audio_stream_buffer (out_stream);
    }

  display ();
}

void
motionFunc (int x, int y)
{

  double my;
  x -= SCREEN_W / 2;
  y -= SCREEN_H / 2;

  if(ego)
  {
  heading = x / 50.0;
    roulette_sim->horizon = -y / 2.0;
  }
  else
  {
	heading = atan2(x,y);
	roulette_sim->horizon = sqrt(x*x+y*y);
  }
  
  mtilt_angle_x = sin (heading)  ;
  mtilt_angle_z = cos (heading) ;



  if (roulette_sim->horizon < 2)
    roulette_sim->horizon = 2;
  if (roulette_sim->horizon >= roulette_sim->steps)
    roulette_sim->horizon = roulette_sim->steps - 1;
	
	if(mouse_b)
	{
		while(mouse_b);
		jump_to_arrow=1;
	}
}






roulette *
create_roulette (int steps, int branch, int sens)
{
  int i, j;
  int n_balls;
  roulette *sim;
  sim = calloc (sizeof (*sim), 1);
  sim->steps = steps;
  sim->sens = sens;
  sim->branch = branch;
  sim->bearings = calloc (sizeof (*(sim->bearings)), steps);
  sim->horizon = sim->steps - 1;
  for (i = 0; i < steps; i++)
    {
      n_balls = branch * sens;
      sim->bearings[i] = calloc (sizeof (**(sim->bearings)), n_balls);
      for (j = 0; j < n_balls; j++)
	sim->bearings[i][j] = create_ball (0, 0, 0);
    }
  return sim;
}


void
parseScreenSize (char *size)
{
  long x, y;
  char *xIndex;
  x = strtol (size, &xIndex, 10);
  y = strtol (xIndex + 1, NULL, 10);
  if (x != 0 && y != 0)
    {
      screen_w = x;
      screen_h = y;
    }
}

void
parseMonteCarlo (char *size)
{
  long steps, branch;
  char *xIndex;
  steps = strtol (size, &xIndex, 10);
  branch = strtol (xIndex + 1, NULL, 10);
  if (steps != 0 && branch != 0)
    {
      monte_carlo = 1;

      roulette_sim = create_roulette (steps, branch, 1);
    }
}


double *
parseLine (char *line, int *line_length)
{
  char *remainder;
  int i;
  double *line_vals = calloc (sizeof (*line_vals), 200);
  i = 0;
  line_vals[i++] = strtod (line, &remainder);
  while (*remainder && *remainder != '\n')
    line_vals[i++] = strtod (remainder, &remainder);

  *line_length = i - 1;
  return line_vals;
}


void keyFunc(void)
{
double s = 1.0;

if(key[KEY_DOWN])
{
	one_bearing->z-=mtilt_angle_z*2;
	one_bearing->x-=mtilt_angle_x*2;
}
if(key[KEY_UP])
{
	one_bearing->z+=mtilt_angle_z*2;
	one_bearing->x+=mtilt_angle_x*2;
}

if(key[KEY_B])
{
BITMAP *tmp;
surface *stmp;
while(key[KEY_B]);
tmp = bike;
stmp = bike_ground;
bike = map;
bike_ground = ground;
map = tmp;
ground = stmp;
}
if(key[KEY_S])
{
	while(key[KEY_S]);
	shadowing=!shadowing;
	}
	
if(key[KEY_Q])
pos_petrb+=0.01;
if(key[KEY_A] && pos_petrb>0)
pos_petrb-=0.01;

if(key[KEY_E])
step_ptrb+=0.01;
if(key[KEY_D] && step_ptrb>0)
step_ptrb-=0.01;

if(key[KEY_R])
petrb+=0.005;
if(key[KEY_F] && petrb>0)
petrb-=0.005;

if(key[KEY_M])
{	
	while(key[KEY_M]);
	if(drop_mode)
	{
		fprintf(marker_file, "%s,%d,%d\n", marker_wave_names[cur_wave], (int)one_bearing->x, (int)one_bearing->z);
		fflush(marker_file);
		wave_x[cur_wave] = one_bearing->x;
		wave_y[cur_wave] = one_bearing->z;
		
		cur_wave++;
		if(cur_wave>=n_waves)
			drop_mode = 0;
	}
	
}

if(key[KEY_X])
{
	while(key[KEY_X]);
	drop_mode = !drop_mode;
	if(drop_mode)
	{
		int i;
		marker_file = fopen("markers.txt", "w");
		cur_wave = 0;
		for(i=0;i<n_waves;i++)
			{
				wave_x[i] = 0;
				wave_y[i] = 0;
			}
	}
	else
		marker_file = NULL;
}
	if(key[KEY_C])
	{
	while(key[KEY_C]);
	ego=!ego;
	egopos = !egopos;
	
	}
	if(key[KEY_T])
{
	while(key[KEY_T]);
	trans_mode=!trans_mode;
	}
}

void load_maps(void)
{
	map = load_bitmap("map_walls.bmp", NULL);
	bike = load_bitmap("map_bike.bmp", NULL);
	shadow = load_bitmap("shadow.bmp", NULL);
	if(!map || !shadow || map->w!=shadow->w || map->h!=shadow->h)
		{
			printf("Maps don't match!\n");
			exit(-1);
		}
	ground = create_surface(map, shadow);
	bike_ground = create_surface(bike, shadow);
}

int
main (int argc, char **argv)
{
  int i;
  int buf_size, quality;

  long rest_time;

  allegro_init ();
  install_timer ();
  install_keyboard ();
  install_mouse ();
  if (install_sound (DIGI_AUTODETECT, MIDI_NONE, NULL) != 0)
    {
      printf ("Could not initialize sound...\n");
      exit (-1);
    }
  set_color_depth (32);
  if (set_gfx_mode (GFX_DIRECTX_WIN, 800, 600, 800, 600) != 0)
    {
      set_color_depth (24);
      if (set_gfx_mode (GFX_DIRECTX_WIN, 800, 600, 800, 600) != 0)
	{
	  printf ("Could not create graphics window...\n");
	  exit (-1);
	}
    }


  screen_buf = create_bitmap (SCREEN_W, SCREEN_H);


  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] == '-' && argv[i][1] == 's' && i < (argc - 1))
	parseScreenSize (argv[i + 1]);

      if (argv[i][0] == '-' && argv[i][1] == 'm' && i < (argc - 1))
	parseMonteCarlo (argv[i + 1]);



    }

  
  load_maps();



  one_bearing = create_ball (0, 0, 0);
  one_bearing->x = 7;
  one_bearing->z = 7;

  init_grains ();
  load_waveforms ();
  

  out_stream =
    play_audio_stream (BUFSIZE, 16, FALSE, SAMPLE_RATE / QUALITY, 255, 128);
  dbl_buf = malloc (sizeof (char) * BUFSIZE * 2);
  set_buffer_length (256);

  buf = get_audio_stream_buffer (out_stream);
  sscale = SCREEN_W / (double) (ground->s_x * 3.0);
  while (!key[KEY_ESC])
    {
      timerFunc ();
	  keyFunc();
    }

  stop_audio_stream (out_stream);

  return EXIT_SUCCESS;
}

END_OF_MAIN ();
